import json
import os
import sys

def get_config() -> dict:
    current_dir = os.getcwd()
    config = {}
    if os.path.exists("nozy_config.json"):
        with open("config.json", "r") as f:
            config = json.load(f)
    config = {
        'package_path': 'nozy/packages',
        'model_paths': ['models'],
        'workflow_path': 'nozy/workflows',
        'output_path': 'nozy/outputs',
        'HF_HOME': './models',
        'HF_HUB_OFFLINE': 'False',
        'ignore': ['node_modules', 'venv', '.env', '.git', '__pycache__', 'dist', '.DS_Store', 'nozyio.egg-info', 'nozyio'],
        **config,
    }
    if isinstance(config.get('package_path'), str):
        config['package_path'] = os.path.join(current_dir, config['package_path'])
    if isinstance(config.get('model_paths'), str):
        config['model_paths'] = os.path.join(current_dir, config['model_paths'])
    if isinstance(config.get('model_paths'), list):
        config['model_paths'] = [os.path.join(current_dir, path) for path in config['model_paths']]
    if isinstance(config.get('workflow_path'), str):
        config['workflow_path'] = os.path.join(current_dir, config['workflow_path'])
    if isinstance(config.get('output_path'), str):
        config['output_path'] = os.path.join(current_dir, config['output_path'])

    sys.path.append(config['package_path'])
    os.environ['HF_HOME'] = config['HF_HOME']
    os.environ['HF_HUB_OFFLINE'] = config['HF_HUB_OFFLINE']
    return config

def get_root_dir():
    return os.getcwd()

def should_ignore(path: str) -> bool:
    if path.startswith('.'):
        return True
    return any(path.startswith(ignore) for ignore in config['ignore'])

config = get_config()

nz_config = config


import ast
import importlib
import json
import os
import traceback
def is_serializable(value):
    try:
        json.dumps(value)
        return True
    except:
        return False

def infer_type_from_default(default):
    if default is None:
        return None
    elif isinstance(default, bool):
        return 'bool'
    elif isinstance(default, int):
        return 'int'
    elif isinstance(default, float):
        return 'float'
    elif isinstance(default, str):
        return 'str'
    elif isinstance(default, list):
        return 'list'
    elif isinstance(default, dict):
        return 'dict'
    else:
        return None

import ast
import json
from typing import Literal

def is_serializable(value):
    try:
        json.dumps(value)
        return True
    except:
        return False

def infer_type_from_default(default):
    if default is None:
        return None
    elif isinstance(default, bool):
        return 'bool'
    elif isinstance(default, int):
        return 'int'
    elif isinstance(default, float):
        return 'float'
    elif isinstance(default, str):
        return 'str'
    elif isinstance(default, list):
        return 'list'
    elif isinstance(default, dict):
        return 'dict'
    else:
        return None

def process_annotation(annotation, import_aliases):
    if annotation is not None:
        if (isinstance(annotation, ast.Subscript) and
            isinstance(annotation.value, ast.Name) and
            annotation.value.id == 'Union'):
            # Extract the types inside Union
            if isinstance(annotation.slice, ast.Tuple):
                # Multiple types
                types = []
                for elt in annotation.slice.elts:
                    type_name = ast.unparse(elt)
                    # Resolve alias to real class name and extract class name
                    resolved_name = import_aliases.get(type_name, type_name)
                    types.append(resolved_name.split('.')[-1])
                # return types
                return 'Union'
            else:
                # Single type
                type_name = ast.unparse(annotation.slice)
                resolved_name = import_aliases.get(type_name, type_name)
                return [resolved_name.split('.')[-1]]
        # Check if the annotation is a Literal
        elif (isinstance(annotation, ast.Subscript) and
            isinstance(annotation.value, ast.Name) and
            annotation.value.id == 'Literal'):
            # Extract the values inside Literal
            if isinstance(annotation.slice, ast.Tuple):
                # Multiple values
                literal_values = []
                for elt in annotation.slice.elts:
                    try:
                        value = ast.literal_eval(elt)
                        literal_values.append(value)
                    except:
                        pass  # Cannot evaluate
                return literal_values
            else:
                # Single value
                try:
                    value = ast.literal_eval(annotation.slice)
                    return [value]
                except:
                    pass  # Cannot evaluate
        else:
            # For other annotations, unparse to get the string representation
            annotation_str = ast.unparse(annotation)
            # Check if the annotation is an alias and map it to the real class name
            full_name = import_aliases.get(annotation_str, annotation_str)
            # Extract only the class name
            return full_name.split('.')[-1]
    return None

def extract_function_info_from_ast(function_node, module_name, nozy_node_def = {}, import_aliases = {}):
    # Get the function name
    function_name = function_node.name

    # Search for NOZY_NODE_DEF assignment within the function body
    for node in function_node.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == 'NOZY_NODE_DEF':
                    try:
                        # Evaluate the assigned value
                        nozy_node_def = ast.literal_eval(node.value)
                    except:
                        pass
                    break

    if not isinstance(nozy_node_def, dict):
        nozy_node_def = {}

    extra_input_info = nozy_node_def.get('input', nozy_node_def.get('inputs', {}))
    if not isinstance(extra_input_info, dict):
        extra_input_info = {}

    # Now extract parameters, defaults, annotations, etc.
    ast_args = []
    args = function_node.args

    # Process positional arguments
    num_args = len(args.args)
    num_defaults = len(args.defaults)
    num_non_defaults = num_args - num_defaults

    for i, arg in enumerate(args.args):
        param_name = arg.arg
        param_annotation = arg.annotation

        if i >= num_non_defaults:
            # Argument has default
            default_index = i - num_non_defaults
            default_node = args.defaults[default_index]
            try:
                default_value = ast.literal_eval(default_node)
            except:
                default_value = None
            optional = True
        else:
            default_value = None
            optional = False

        type_annotation = process_annotation(param_annotation, import_aliases)
        inferred_type = infer_type_from_default(default_value)
        param_type = type_annotation or inferred_type or "any"

        ast_args.append({
            "id": str(param_name),
            "io_type": "input",
            "default": default_value if is_serializable(default_value) else None,
            "optional": optional,
            "name": param_name,
            "type": param_type,
            **extra_input_info.get(param_name, {})
        })

    # Process vararg (*args)
    if args.vararg:
        param_name = args.vararg.arg
        param_annotation = args.vararg.annotation
        type_annotation = process_annotation(param_annotation, import_aliases)
        param_type = type_annotation or "any"

        ast_args.append({
            "id": str(param_name),
            "io_type": "input",
            "default": None,
            "optional": False,
            "name": "*" + param_name,
            "type": param_type,
            **extra_input_info.get(param_name, {})
        })

    # Process keyword-only arguments
    for i, kwarg in enumerate(args.kwonlyargs):
        param_name = kwarg.arg
        param_annotation = kwarg.annotation
        default_node = args.kw_defaults[i]

        if default_node is not None:
            try:
                default_value = ast.literal_eval(default_node)
            except:
                default_value = None
            optional = True
        else:
            default_value = None
            optional = False

        type_annotation = process_annotation(param_annotation, import_aliases)
        inferred_type = infer_type_from_default(default_value)
        param_type = type_annotation or inferred_type or "any"

        ast_args.append({
            "id": str(param_name),
            "io_type": "input",
            "default": default_value if is_serializable(default_value) else None,
            "optional": optional,
            "name": param_name,
            "type": param_type,
            **extra_input_info.get(param_name, {})
        })

    # Process kwarg (**kwargs)
    if args.kwarg:
        param_name = args.kwarg.arg
        param_annotation = args.kwarg.annotation
        type_annotation = process_annotation(param_annotation, import_aliases)
        param_type = type_annotation or "any"

        ast_args.append({
            "id": str(param_name),
            "io_type": "input",
            "default": None,
            "optional": False,
            "name": "**" + param_name,
            "type": param_type,
            **extra_input_info.get(param_name, {})
        })

    # Handle outputs (return type annotation)
    outputs = []
    if function_node.returns is not None:
        return_annotation = process_annotation(function_node.returns, import_aliases)
        if isinstance(return_annotation, list):
            # Multiple output types
            for index, output_type in enumerate(return_annotation):
                outputs.append({
                    "id": index,
                    "io_type": "output",
                    "type": output_type
                })
        else:
            outputs.append({
                "id": "0",
                "io_type": "output",
                "type": return_annotation
            })
    else:
        outputs.append({
            "id": "0",
            "io_type": "output",
            "type": "any"
        })
    extra_output_info = nozy_node_def.get('output', nozy_node_def.get('outputs', None))
    if extra_output_info is not None:
        for index,extra_output in enumerate(extra_output_info):
            extra_output['id'] = str(index)
            extra_output['io_type'] = 'output'
        outputs = extra_output_info

    # Create the assign AST node
    if len(outputs) == 1:
        targets = outputs
    else:
        targets = [{
            "node_type": "Tuple",
            "elts": outputs,
            "ctx": {
                "node_type": "Store"
            }
        }]

    assign_ast_node = {
        "node_type": "Assign",
        "value": {
            "node_type": "Call",
            "func": {
                "node_type": "Name",
                "id": function_name,
                "ctx": {
                    "node_type": "Load"
                }
            },
            "args": ast_args,
            "keywords": []
        },
        "targets": targets
    }

    # Create the import AST node
    import_ast_node = {
        "node_type": "ImportFrom",
        "module": module_name,
        "names": [{
            "node_type": "alias",
            "name": function_name,
            "asname": None
        }],
        "level": 0
    }

    return {
        "type": "function",
        "node_title": nozy_node_def.get('node_title', function_name),
        "description": nozy_node_def.get('description', None),
        "name": function_name,
        "module": module_name,
        "input": ast_args,
        "output": outputs,
        "astNodes": [import_ast_node, assign_ast_node]
    }

def build_import_aliases(tree):
    import_aliases = {}
    for node in ast.walk(tree):
        if isinstance(node, ast.ImportFrom):
            module = node.module
            for alias in node.names:
                import_aliases[alias.asname or alias.name] = f"{module}.{alias.name}"
        elif isinstance(node, ast.Import):
            for alias in node.names:
                import_aliases[alias.asname or alias.name] = alias.name
    return import_aliases

def parse_python_file_dict(file_path, module_name) -> dict[str, dict]:
    """Parse a Python file and extract function information."""
    with open(file_path, "r") as file:
        source_code = file.read()

    tree = ast.parse(source_code)
    functions = {}
    nozy_node_defs = {}
    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Attribute) and target.attr == 'NOZY_NODE_DEF':
                    if isinstance(target.value, ast.Name):
                        # is NOZY_NODE_DEF assign
                        function_name = target.value.id
                        try:
                            nozy_node_def = ast.literal_eval(node.value)
                            if isinstance(nozy_node_def, dict):
                                nozy_node_defs[function_name] = nozy_node_def
                        except Exception as e:
                            print('❌error parsing NOZY_NODE_DEF', e, traceback.format_exc())
                            
    import_aliases = build_import_aliases(tree)
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            # Extract function information
            func_info = extract_function_info_from_ast(node, module_name, nozy_node_defs.get(node.name, {}), import_aliases)
            functions[node.name] = func_info

    return functions

def parse_python_file(file_path, module_name) -> list[dict]:
    details_dict = parse_python_file_dict(file_path, module_name)
    return list(details_dict.values())

def get_function_details_by_ast(module_name, function_name):
    # Find the module specification
    spec = importlib.util.find_spec(module_name)
    if spec is None:
        raise ImportError(f"Module {module_name} not found")
    module_path = spec.origin
    if module_path is None:
        raise ImportError(f"Module {module_name} does not have a file path")
    
    file_path = os.path.abspath(module_path)
    funcitons = parse_python_file_dict(file_path, module_name)
    return funcitons.get(function_name, None)

if __name__ == "__main__":
    # Example usage
    file_path = "path/to/your/module.py"
    module_name = "module"
    functions_info = parse_python_file(file_path, module_name)
    for func in functions_info:
        print(func)


import importlib
import inspect
import json
import os
import platform
import stat
import subprocess
import re
import traceback
from .scan_modules_ast import get_function_details_by_ast
from .scan_modules import extract_function_details

# import pkg_resources

# def get_pip_packages_list() -> list:
#     packages = []
#     for dist in pkg_resources.working_set:
#         packages.append({
#             'name': dist.project_name,
#             'version': dist.version
#         })
#     return packages

def get_ripgrep_binary():
    system = platform.system().lower()

    if system == "linux":
        return os.path.join(os.path.dirname(__file__), 'bin', 'linux', 'rg')
    elif system == "darwin":
        return os.path.join(os.path.dirname(__file__), 'bin', 'mac', 'rg')
    elif system == "windows":
        return os.path.join(os.path.dirname(__file__), 'bin', 'windows', 'rg.exe')
    else:
        raise OSError(f"Unsupported operating system: {system}")

def ensure_executable_permissions(binary_path):
    """Ensure the binary has executable permissions (only on Unix-like systems)."""
    system = platform.system().lower()

    if system in ['linux', 'darwin']:  # Only check on Linux and macOS
        if not os.access(binary_path, os.X_OK):
            # If the file is not executable, add the executable bit
            st = os.stat(binary_path)
            os.chmod(binary_path, st.st_mode | stat.S_IEXEC)
    else:
        # On Windows, we assume the binary (e.g., .exe) doesn't need this check
        pass

binary_path = get_ripgrep_binary()
ensure_executable_permissions(binary_path)

def search_codebase(search_term):
    def parse_match(line):
        # Regex pattern to capture file path, line number, and definition type and name
        pattern = r'^(?P<file_path>.*?):\d+:\s*(?P<type>class|def)\s+(?P<name>\w+)'
        # pattern = r'^(?P<file_path>.*?):\d+:\s*def\s+(?P<name>\w+)\s*\(' # search module level functions only
        match = re.match(pattern, line)
        if match:
            file_path = os.path.relpath(match.group("file_path"))
            return {
                "name": match.group("name"),
                "type": "class" if match.group("type") == "class" else "function",
                "file_path": file_path
            }
        return None

    rg_command = [
        binary_path, '-n', '--no-heading', '--color', 'never',
        rf'^\s*(class|def)\s+\w*{re.escape(search_term)}\w*',  # regex to find class and function definitions containing the search term
        '.',
        '--glob', '*.py',  # Include only .py files
    ]

    result = subprocess.run(rg_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    MAX_RESULTS = 1000
    if result.returncode == 0:
        matches = result.stdout.splitlines()
        matches = matches[:MAX_RESULTS]

        # Parse each match to extract file_path, function name, type
        search_results = []
        for match in matches:
            parsed = parse_match(match)
            if parsed:
                search_results.append(parsed)
                if len(search_results) >= MAX_RESULTS:
                    break

        search_results = search_results[:MAX_RESULTS]
        # import modules to extract function details
        for index, result in enumerate(search_results):
            try:
                relative_path = os.path.relpath(result['file_path'])
                module_name = os.path.splitext(relative_path)[0].replace(os.path.sep, '.')
                # module = importlib.import_module(module_name)
                # if hasattr(module, result['name']):
                #     function_obj = getattr(module, result['name'])
                #     func_info = extract_function_details(function_obj, module.__name__)
                #     search_results[index] = func_info
                # else:
                #     # is class or instance method
                #     search_results[index] = None
                search_results[index] = get_function_details_by_ast(module_name, result['name'])
            except Exception as e:
                search_results[index] = None
                print(e, traceback.format_exc())

        return search_results
    else:
        return []

def typeahead_search():
    search_term = input("Start typing class/function name: ")
    while search_term:
        # Fetch the results from the codebase based on the search term
        matches = search_codebase(search_term)
        print("\n".join(json.dumps(match) for match in matches))
        
        search_term = input("Start typing class/function name: ")

if __name__ == '__main__':
    typeahead_search()



import os

from .config_utils import config, get_root_dir, should_ignore


def list_files(path, extensions=None):
    if not os.path.exists(path):
        return []
    children = []
    ignore_files = config.get('ignore', [])
    for file in os.listdir(path):
        ext = file.split(".")[-1]
        abs_path = os.path.join(path, file)
        rel_path = os.path.relpath(abs_path, get_root_dir())
        if should_ignore(file):
            continue

        if os.path.isdir(abs_path):
            children.append({
                "type": "folder",
                "name": file,
                "path": abs_path,
                "rel_path": rel_path
            })
        else:
            if extensions and ext not in extensions and f'.{ext}' not in extensions:
                continue
            children.append({
                "type": "file",
                "name": file,
                "path": abs_path,
                "rel_path": rel_path
            })
    return children



from typing import Literal
import torch

device: Literal['cpu', 'mps', 'cuda'] = 'cpu'

if torch.cuda.is_available():
    device = 'cuda'
elif torch.backends.mps.is_available():
    device = 'mps'
else:
    device = 'cpu'



import json
import math
import os
import subprocess
import tempfile
import traceback
from .config_utils import config

def get_username_repo_from_url(url: str):
    if not url.startswith("https://github.com/"):
        raise ValueError("Invalid URL")
    if url.endswith(".git"):
        url = url[:-4]
    if url.endswith("/"):
        url = url[:-1]
    username, repo = url.split("/")[-2:]
    return username, repo

def install_node_package(url: str):
    package_path = config['package_path']
    username, repo = get_username_repo_from_url(url)
    repo_path = os.path.join(package_path, repo)
    if os.path.exists(repo_path):
        raise Exception(f"Package {repo} already installed")
    
    try:
        subprocess.run(["git", "clone", url, repo_path], check=True)
    except subprocess.CalledProcessError as e:
        print(e, traceback.format_exc())
        raise Exception(f"Failed to clone repository: {e}")
    
    requirements_path = os.path.join(repo_path, "requirements.txt")
    if os.path.exists(requirements_path):
        try:
            subprocess.run(["pip", "install", "-r", requirements_path], check=True)
        except subprocess.CalledProcessError as e:
            print(e, traceback.format_exc())
            raise Exception(f"Failed to install requirements: {e}")
    
    install_py_path = os.path.join(repo_path, "install.py")
    if os.path.exists(install_py_path):
        try:
            subprocess.run(["python", install_py_path], check=True)
        except subprocess.CalledProcessError as e:
            print(e, traceback.format_exc())
            raise Exception(f"Failed to run install.py: {e}")

def list_community_node_packages():
    packages = []
    community_packages_list = os.path.join(os.path.dirname(__file__), "community_packages.json")
    with open(community_packages_list, "r") as f:
        packages = json.load(f)
    return packages
    


import asyncio
import json
import sys
import tempfile
from aiohttp import web
import os
from .node_package import list_community_node_packages, install_node_package
from .code_to_graph import code_to_graph
from .ast_execution import execute_graph, graph_to_code
from .file_picker import list_files
from .scan_modules import refresh_node_def, scan_directory
from .config_utils import config, get_root_dir
from .search_codebase import search_codebase
from .websocket_manager import websocket_manager
from .job_queue_manager import on_cleanup, process_queue, run_in_executor

WEB_PATH = os.path.join(os.path.dirname(__file__), 'web/dist')

def endpoint(path, method='GET'):
    def decorator(func):
        func.route_info = {'path': path, 'methods': [method]}
        return func
    return decorator

@endpoint('/queue_status')
async def handle_queue_status(request):
    task_queue = request.app['task_queue']
    return web.json_response({
        'queue_size': task_queue.qsize(),
        'tasks_pending': not task_queue.empty()
    })

@endpoint('/get_os_sep')
async def handle_get_os_sep(request):
    return web.Response(text=os.sep)

@endpoint('/code_to_graph')
async def handle_code_to_graph(request):
    rel_path = request.rel_url.query['path']
    abs_path = os.path.join(config['package_path'], rel_path)
    return web.json_response(code_to_graph(abs_path))

@endpoint('/queue_job', method='POST')
async def handle_queue_job(request):
    body = await request.json()
    task_queue = request.app['task_queue']
    await task_queue.put(body)  # Add the task to the queue
    print('Job queued')
    return web.json_response({'status': 'queued'})

@endpoint('/run_job', method='POST')
async def handle_run_job(request):
    body = await request.json()
    # await run_in_executor(execute_graph, body)
    await asyncio.to_thread(execute_graph, body)
    return web.json_response({'status': 'ok'})

@endpoint('/list_package_children')
async def handle_scan_directories(request):
    rel_path = request.rel_url.query['path']
    blacklist = config.get('blacklist', [])
    abs_path = os.path.join(get_root_dir(), rel_path)
    children = scan_directory(abs_path)
    return web.json_response(children)

@endpoint('/preview_image')
async def handle_preview_image(request):
    path = request.rel_url.query['path']
    download = request.rel_url.query.get('download', False)
    abs_path = os.path.join(get_root_dir(), path)

    if not os.path.exists(abs_path) or not os.path.isfile(abs_path):
        return web.Response(status=404, text="Image not found")

    with open(abs_path, 'rb') as image_file:
        image_data = image_file.read()
    if download:
        return web.Response(
        body=image_data, 
        content_type='image/jpeg',
        headers={
            'Content-Disposition': f'attachment; filename="{os.path.basename(abs_path)}"'
        })
    else:
        return web.Response(body=image_data, content_type='image/jpeg')  # Adjust content_type as needed


@endpoint('/file_picker/list_files', method='POST')
async def handle_list_files(request):
    body = await request.json()
    path = body['path']
    extensions = body.get('extensions', None)
    abs_path = os.path.join(get_root_dir(), path)
    children = list_files(abs_path, extensions)
    return web.json_response({
        'path': abs_path,
        'items': children
    })

@endpoint('/graph_to_code', method='POST')
async def handle_graph_to_code(request):
    graph = await request.json()
    code = graph_to_code(graph)
    return web.json_response(code)

@endpoint('/ws')
async def websocket_handler(request):
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    print('WebSocket connection established')
    await websocket_manager.set_connection(ws)


    async for msg in ws:
        if msg.type == web.WSMsgType.TEXT:
            # Handle incoming messages if needed
            print(f"Received message: {msg.data}")
            await websocket_manager.send_message(f"Echo: {msg.data}")
        elif msg.type == web.WSMsgType.ERROR:
            print(f"WebSocket connection closed with exception {ws.exception()}")

    return ws

@endpoint('/search_functions')
async def handle_search_functions(request):
    query = request.rel_url.query['query']
    results = await asyncio.to_thread(search_codebase, query)
    return web.json_response(results)

@endpoint('/workflow/save', method='POST')
async def handle_workflow_save(request):
    body = await request.json()
    graph = body['graph']
    path = body.get('path', '')
    new = body.get('new', False)
    abs_path = os.path.join(config['workflow_path'], path)
    def save_workflow(graph, new):
        if new and os.path.exists(abs_path):
            raise Exception(f'File {abs_path} already exists, please change the name')
        dir = os.path.dirname(abs_path)
        if not os.path.exists(dir):
            os.makedirs(dir)
        with open(abs_path, 'w', encoding='utf-8') as f:  # Specify UTF-8 encoding
            f.write(json.dumps(graph, ensure_ascii=False))  # Ensure non-ASCII characters are preserved
    try:
        await asyncio.to_thread(save_workflow, graph, new)
    except Exception as e:
        return web.json_response({'error': str(e)}, status=400)
    return web.json_response({'path': path, 'root_rel': os.path.relpath(abs_path, get_root_dir())})

@endpoint('/workflow/rename', method='POST')
async def handle_workflow_rename(request):
    body = await request.json()
    path = body['path']
    new_name = body['new_name']
    abs_path = os.path.join(config['workflow_path'], path)
    new_abs_path = os.path.join(os.path.dirname(abs_path), new_name)
    if os.path.exists(new_abs_path):
        return web.json_response({'error': f'{new_name} already exists, please change the name'}, status=400)
    os.rename(abs_path, new_abs_path)
    return web.json_response({'path': os.path.relpath(new_abs_path, config['workflow_path']), 'root_rel': os.path.relpath(new_abs_path, get_root_dir())})

@endpoint('/workflow/list')
async def handle_workflow_list(request):
    path = request.rel_url.query['path']
    abs_path = os.path.join(config['workflow_path'], path)
    children = []
    for file in os.listdir(abs_path):
        ext = file.split(".")[-1]
        if ext != 'json':
            continue
        rel_path = os.path.relpath(abs_path, config['workflow_path'])
        file_path = os.path.join(abs_path, file)

        if os.path.isdir(file_path):
            children.append({
                "type": "folder",
                "name": file,
                "path": file_path,
                "rel_path": os.path.relpath(file_path, config['workflow_path'])
            })
        else:
            children.append({
                "type": "file",
                "name": file,
                "path": file_path,
                "rel_path": os.path.relpath(file_path, config['workflow_path'])
            })
    return web.json_response(children)

@endpoint('/workflow/get')
async def handle_workflow_get(request):
    path = request.rel_url.query['path']
    abs_path = os.path.join(config['workflow_path'], path)
    with open(abs_path, 'r', encoding='utf-8') as f:
        graph = json.load(f)
    return web.json_response(graph)

@endpoint('/workflow/get_default')
async def handle_workflow_get_default(request):
    abs_path = config['workflow_path']
    # find the first json file
    for file in os.listdir(abs_path):
        if file.endswith('.json'):
            return web.json_response(file)
    return web.json_response(None)

@endpoint('/refresh_node_def', method='POST')
async def handle_refresh_node_def(request):
    graph = await request.json()
    new_graph = refresh_node_def(graph)
    return web.json_response(new_graph)

@endpoint('/package/list_community')
async def handle_list_community_packages(request):
    packages = list_community_node_packages()
    return web.json_response(packages)

@endpoint('/package/install', method='POST')
async def handle_install_node_package(request):
    body = await request.json()
    url = body['url']
    try:
        await asyncio.to_thread(install_node_package, url)
    except Exception as e:
        return web.json_response({'error': str(e)}, status=400)
    return web.json_response({'status': 'ok'})

@endpoint('/')
async def handle_index(request):
    print('handle_index')
    if os.path.exists(os.path.join(WEB_PATH, 'index.html')):
        print('index.html exists')
        return web.FileResponse(os.path.join(WEB_PATH, 'index.html'))
    else:
        print(f'{os.path.join(WEB_PATH, "index.html")} does not exist')
        return web.Response(status=404)

# Middleware to handle CORS
@web.middleware
async def cors_middleware(request, handler):
    if request.method == 'OPTIONS':
        # Handle preflight CORS requests
        response = web.Response(status=204)
    else:
        response = await handler(request)
    response.headers['Access-Control-Allow-Origin'] = '*'
    return response

async def on_startup(app):
    app['task_queue'] = asyncio.Queue()
    # Start the background task to process the queue
    app['process_queue_task'] = asyncio.create_task(process_queue(app))
    websocket_manager.loop = asyncio.get_running_loop()
    
def start_server():
    allow_cors = '--allow-cors' in sys.argv
    port = int(sys.argv[sys.argv.index('--port') + 1]) if '--port' in sys.argv else 7070
    listen = '--listen' in sys.argv
    host = '0.0.0.0' if listen else '127.0.0.1'

    if allow_cors:
        print('🧪allow cors=', allow_cors)
    app = web.Application(middlewares=[cors_middleware] if allow_cors else [])
    temp_dir = tempfile.gettempdir()
    print("System's temporary directory:", temp_dir)
    # Register routes using the endpoint decorator
    for name, func in globals().items():
        if callable(func) and hasattr(func, 'route_info'):
            route_info = func.route_info
            for method in route_info['methods']:
                app.router.add_route(method, route_info['path'], func)
    
    if os.path.exists(WEB_PATH):
        app.router.add_static('/', WEB_PATH, show_index=True)
    
    app.on_startup.append(on_startup)
    # app.on_cleanup.append(on_cleanup)
    
    # add the current directory to the python path
    sys.path.append(os.getcwd())

    # print('python search path', sys.path)
    web.run_app(app, host=host, port=port)

if __name__ == "__main__":
    print("Starting aiohttp server...")
    start_server()
